@@ -18,6 +18,7 @@ from awkward._nplikes.numpy_like import IndexType, NumpyMetadata
 from awkward._nplikes.placeholder import PlaceholderArray
 from awkward._nplikes.shape import ShapeItem, unknown_length
 from awkward._nplikes.typetracer import MaybeNone, TypeTracer
+from awkward._nplikes.virtual import VirtualArray, materialize_if_virtual
 from awkward._parameters import (
     parameters_intersect,
 )
@@ -131,7 +132,7 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
                 f"{type(self).__name__} 'valid_when' must be boolean, not {valid_when!r}"
             )
         if (
-            content.backend.index_nplike.known_data
+            content.backend.nplike.known_data
             and mask.length is not unknown_length
             and content.length is not unknown_length
             and mask.length > content.length
@@ -140,7 +141,7 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
                 f"{type(self).__name__} len(mask) ({mask.length}) must be <= len(content) ({content.length})"
             )
 
-        assert mask.nplike is content.backend.index_nplike
+        assert mask.nplike is content.backend.nplike
 
         self._mask = mask
         self._content = content
@@ -186,7 +187,7 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
     ):
         if content.is_union or content.is_indexed or content.is_option:
             backend = content.backend
-            index = ak.index.Index64.empty(mask.length, nplike=backend.index_nplike)
+            index = ak.index.Index64.empty(mask.length, nplike=backend.nplike)
             backend.maybe_kernel_error(
                 backend[
                     "awkward_ByteMaskedArray_toIndexedOptionArray",
@@ -238,7 +239,7 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
         assert isinstance(form, self.form_cls)
         key = getkey(self, form, "mask")
         container[key] = ak._util.native_to_byteorder(
-            self._mask.raw(backend.index_nplike), byteorder
+            self._mask.raw(backend.nplike), byteorder
         )
         self._content._to_buffers(form.content, getkey, container, backend, byteorder)
 
@@ -281,7 +282,7 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
         out = [indent, pre, "<ByteMaskedArray valid_when="]
         out.append(repr(json.dumps(self._valid_when)))
         out.append(" len=")
-        out.append(repr(str(self.length)))
+        out.append(repr(str(ak._util.maybe_length_of(self))))
         out.append(">")
         out.extend(self._repr_extra(indent + "    "))
         out.append("\n")
@@ -292,12 +293,10 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
         return "".join(out)
 
     def to_IndexedOptionArray64(self) -> IndexedOptionArray:
-        index = ak.index.Index64.empty(
-            self._mask.length, nplike=self._backend.index_nplike
-        )
+        index = ak.index.Index64.empty(self._mask.length, nplike=self._backend.nplike)
         assert (
-            index.nplike is self._backend.index_nplike
-            and self._mask.nplike is self._backend.index_nplike
+            index.nplike is self._backend.nplike
+            and self._mask.nplike is self._backend.nplike
         )
         self._backend.maybe_kernel_error(
             self._backend[
@@ -321,11 +320,9 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
         else:
             return ByteMaskedArray(
                 ak.index.Index8(
-                    self._backend.index_nplike.astype(
-                        self._backend.index_nplike.logical_not(
-                            self._backend.index_nplike.astype(
-                                self._mask.data, dtype=np.bool_
-                            )
+                    self._backend.nplike.astype(
+                        self._backend.nplike.logical_not(
+                            self._backend.nplike.astype(self._mask.data, dtype=np.bool_)
                         ),
                         dtype=np.int8,
                     )
@@ -339,7 +336,7 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
         if not self._backend.nplike.known_data:
             self._touch_data(recursive=False)
             if self._backend.nplike.known_data:
-                excess_length = int(math.ceil(self.length / 8.0))
+                excess_length = math.ceil(self.length / 8.0)
             else:
                 excess_length = unknown_length
             return ak.contents.BitMaskedArray(
@@ -356,7 +353,7 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
         else:
             bit_order = "little" if lsb_order else "big"
             bytemask = self.mask_as_bool(valid_when).view(np.uint8)
-            bitmask = self.backend.index_nplike.packbits(bytemask, bitorder=bit_order)
+            bitmask = self.backend.nplike.packbits(bytemask, bitorder=bit_order)
 
             return ak.contents.BitMaskedArray(
                 ak.index.IndexU8(bitmask),
@@ -372,9 +369,9 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
             valid_when = self._valid_when
 
         if valid_when == self._valid_when:
-            return self._mask.raw(self._backend.index_nplike) != 0
+            return self._mask.raw(self._backend.nplike) != 0
         else:
-            return self._mask.raw(self._backend.index_nplike) != 1
+            return self._mask.raw(self._backend.nplike) != 1
 
     def _getitem_nothing(self):
         return self._content._getitem_range(0, 0)
@@ -384,6 +381,15 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
             return True
         return self._content._is_getitem_at_placeholder()
 
+    def _is_getitem_at_virtual(self) -> bool:
+        is_virtual = (
+            isinstance(self._mask.data, VirtualArray)
+            and not self._mask.data.is_materialized
+        )
+        if is_virtual:
+            return True
+        return self._content._is_getitem_at_virtual()
+
     def _getitem_at(self, where: IndexType):
         if not self._backend.nplike.known_data:
             self._touch_data(recursive=False)
@@ -446,11 +452,11 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
         )
 
     def _nextcarry_outindex(self) -> tuple[int, ak.index.Index64, ak.index.Index64]:
-        _numnull = ak.index.Index64.empty(1, nplike=self._backend.index_nplike)
+        _numnull = ak.index.Index64.empty(1, nplike=self._backend.nplike)
 
         assert (
-            _numnull.nplike is self._backend.index_nplike
-            and self._mask.nplike is self._backend.index_nplike
+            _numnull.nplike is self._backend.nplike
+            and self._mask.nplike is self._backend.nplike
         )
         self._backend.maybe_kernel_error(
             self._backend[
@@ -464,18 +470,16 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
                 self._valid_when,
             )
         )
-        numnull = self._backend.index_nplike.index_as_shape_item(_numnull[0])
+        numnull = self._backend.nplike.index_as_shape_item(_numnull[0])
         nextcarry = ak.index.Index64.empty(
             self.length - numnull,
-            nplike=self._backend.index_nplike,
-        )
-        outindex = ak.index.Index64.empty(
-            self.length, nplike=self._backend.index_nplike
+            nplike=self._backend.nplike,
         )
+        outindex = ak.index.Index64.empty(self.length, nplike=self._backend.nplike)
         assert (
-            nextcarry.nplike is self._backend.index_nplike
-            and outindex.nplike is self._backend.index_nplike
-            and self._mask.nplike is self._backend.index_nplike
+            nextcarry.nplike is self._backend.nplike
+            and outindex.nplike is self._backend.nplike
+            and self._mask.nplike is self._backend.nplike
         )
         self._backend.maybe_kernel_error(
             self._backend[
@@ -511,17 +515,17 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
 
         reducedstarts = ak.index.Index64.empty(
             self.length - numnull,
-            nplike=self._backend.index_nplike,
+            nplike=self._backend.nplike,
         )
         reducedstops = ak.index.Index64.empty(
             self.length - numnull,
-            nplike=self._backend.index_nplike,
+            nplike=self._backend.nplike,
         )
 
         assert (
-            outindex.nplike is self._backend.index_nplike
-            and slicestarts.nplike is self._backend.index_nplike
-            and slicestops.nplike is self._backend.index_nplike
+            outindex.nplike is self._backend.nplike
+            and slicestarts.nplike is self._backend.nplike
+            and slicestops.nplike is self._backend.nplike
             and reducedstarts.nplike is self._backend.nplike
             and reducedstops.nplike is self._backend.nplike
         )
@@ -597,7 +601,7 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
 
     def project(self, mask=None):
         mask_length = self._mask.length
-        _numnull = ak.index.Index64.zeros(1, nplike=self._backend.index_nplike)
+        _numnull = ak.index.Index64.zeros(1, nplike=self._backend.nplike)
 
         if mask is not None:
             if self._backend.nplike.known_data and mask_length != mask.length:
@@ -605,13 +609,11 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
                     f"mask length ({mask.length}) is not equal to {type(self).__name__} length ({mask_length})"
                 )
 
-            nextmask = ak.index.Index8.empty(
-                mask_length, nplike=self._backend.index_nplike
-            )
+            nextmask = ak.index.Index8.empty(mask_length, nplike=self._backend.nplike)
             assert (
-                nextmask.nplike is self._backend.index_nplike
-                and mask.nplike is self._backend.index_nplike
-                and self._mask.nplike is self._backend.index_nplike
+                nextmask.nplike is self._backend.nplike
+                and mask.nplike is self._backend.nplike
+                and self._mask.nplike is self._backend.nplike
             )
             self._backend.maybe_kernel_error(
                 self._backend[
@@ -635,8 +637,8 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
 
         else:
             assert (
-                _numnull.nplike is self._backend.index_nplike
-                and self._mask.nplike is self._backend.index_nplike
+                _numnull.nplike is self._backend.nplike
+                and self._mask.nplike is self._backend.nplike
             )
             self._backend.maybe_kernel_error(
                 self._backend[
@@ -650,13 +652,13 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
                     self._valid_when,
                 )
             )
-            numnull = self._backend.index_nplike.index_as_shape_item(_numnull[0])
+            numnull = self._backend.nplike.index_as_shape_item(_numnull[0])
             nextcarry = ak.index.Index64.empty(
-                mask_length - numnull, nplike=self._backend.index_nplike
+                mask_length - numnull, nplike=self._backend.nplike
             )
             assert (
-                nextcarry.nplike is self._backend.index_nplike
-                and self._mask.nplike is self._backend.index_nplike
+                nextcarry.nplike is self._backend.nplike
+                and self._mask.nplike is self._backend.nplike
             )
             self._backend.maybe_kernel_error(
                 self._backend[
@@ -695,14 +697,14 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
             else:
                 outoffsets = ak.index.Index64.empty(
                     offsets.length + numnull,
-                    nplike=self._backend.index_nplike,
+                    nplike=self._backend.nplike,
                     dtype=np.int64,
                 )
 
                 assert (
-                    outoffsets.nplike is self._backend.index_nplike
-                    and outindex.nplike is self._backend.index_nplike
-                    and offsets.nplike is self._backend.index_nplike
+                    outoffsets.nplike is self._backend.nplike
+                    and outindex.nplike is self._backend.nplike
+                    and offsets.nplike is self._backend.nplike
                 )
                 self._backend.maybe_kernel_error(
                     self._backend[
@@ -742,14 +744,12 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
             for x in others
         ):
             parameters = self._parameters
-            self_length_scalar = self._backend.index_nplike.shape_item_as_index(
-                self.length
-            )
+            self_length_scalar = self._backend.nplike.shape_item_as_index(self.length)
             masks = [self._mask.data[:self_length_scalar]]
             tail_contents = []
             length = 0
             for x in others:
-                length_scalar = self._backend.index_nplike.shape_item_as_index(x.length)
+                length_scalar = self._backend.nplike.shape_item_as_index(x.length)
                 parameters = parameters_intersect(parameters, x._parameters)
                 masks.append(x._mask.data[:length_scalar])
                 tail_contents.append(x._content[:length_scalar])
@@ -846,10 +846,10 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
     ):
         mask_length = self._mask.length
 
-        _numnull = ak.index.Index64.empty(1, nplike=self._backend.index_nplike)
+        _numnull = ak.index.Index64.empty(1, nplike=self._backend.nplike)
         assert (
-            _numnull.nplike is self._backend.index_nplike
-            and self._mask.nplike is self._backend.index_nplike
+            _numnull.nplike is self._backend.nplike
+            and self._mask.nplike is self._backend.nplike
         )
         self._backend.maybe_kernel_error(
             self._backend[
@@ -863,24 +863,18 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
                 self._valid_when,
             )
         )
-        numnull = self._backend.index_nplike.index_as_shape_item(_numnull[0])
+        numnull = self._backend.nplike.index_as_shape_item(_numnull[0])
 
         next_length = mask_length - numnull
-        nextcarry = ak.index.Index64.empty(
-            next_length, nplike=self._backend.index_nplike
-        )
-        nextparents = ak.index.Index64.empty(
-            next_length, nplike=self._backend.index_nplike
-        )
-        outindex = ak.index.Index64.empty(
-            mask_length, nplike=self._backend.index_nplike
-        )
+        nextcarry = ak.index.Index64.empty(next_length, nplike=self._backend.nplike)
+        nextparents = ak.index.Index64.empty(next_length, nplike=self._backend.nplike)
+        outindex = ak.index.Index64.empty(mask_length, nplike=self._backend.nplike)
         assert (
-            nextcarry.nplike is self._backend.index_nplike
-            and nextparents.nplike is self._backend.index_nplike
-            and outindex.nplike is self._backend.index_nplike
-            and self._mask.nplike is self._backend.index_nplike
-            and parents.nplike is self._backend.index_nplike
+            nextcarry.nplike is self._backend.nplike
+            and nextparents.nplike is self._backend.nplike
+            and outindex.nplike is self._backend.nplike
+            and self._mask.nplike is self._backend.nplike
+            and parents.nplike is self._backend.nplike
         )
         self._backend.maybe_kernel_error(
             self._backend[
@@ -905,12 +899,12 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
 
         if reducer.needs_position and (not branch and negaxis == depth):
             nextshifts = ak.index.Index64.empty(
-                next_length, nplike=self._backend.index_nplike
+                next_length, nplike=self._backend.nplike
             )
             if shifts is None:
                 assert (
-                    nextshifts.nplike is self._backend.index_nplike
-                    and self._mask.nplike is self._backend.index_nplike
+                    nextshifts.nplike is self._backend.nplike
+                    and self._mask.nplike is self._backend.nplike
                 )
                 self._backend.maybe_kernel_error(
                     self._backend[
@@ -926,8 +920,8 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
                 )
             else:
                 assert (
-                    nextshifts.nplike is self._backend.index_nplike
-                    and self._mask.nplike is self._backend.index_nplike
+                    nextshifts.nplike is self._backend.nplike
+                    and self._mask.nplike is self._backend.nplike
                 )
                 self._backend.maybe_kernel_error(
                     self._backend[
@@ -977,7 +971,7 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
                 )
 
             outoffsets = ak.index.Index64.empty(
-                starts.length + 1, nplike=self._backend.index_nplike
+                starts.length + 1, nplike=self._backend.nplike
             )
             assert outoffsets.nplike is self._backend.nplike
             self._backend.maybe_kernel_error(
@@ -1013,12 +1007,10 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
             return self._pad_none_axis0(target, clip)
         elif posaxis is not None and posaxis + 1 == depth + 1:
             mask = ak.index.Index8(self.mask_as_bool(valid_when=False))
-            index = ak.index.Index64.empty(
-                mask.length, nplike=self._backend.index_nplike
-            )
+            index = ak.index.Index64.empty(mask.length, nplike=self._backend.nplike)
             assert (
-                index.nplike is self._backend.index_nplike
-                and self._mask.nplike is self._backend.index_nplike
+                index.nplike is self._backend.nplike
+                and self._mask.nplike is self._backend.nplike
             )
             self._backend.maybe_kernel_error(
                 self._backend[
@@ -1065,7 +1057,9 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
         cp = Cupy.instance()._module
 
         assert mask is None  # this class has its own mask
-        m = cp.packbits(cp.asarray(self._mask), bitorder="little")
+        m = cp.packbits(
+            cp.asarray(*materialize_if_virtual(self._mask.data)), bitorder="little"
+        )
         if m.nbytes % 64:
             m = cp.resize(m, ((m.nbytes // 64) + 1) * 64)
         m = cudf.core.buffer.as_buffer(m)
@@ -1195,11 +1189,26 @@ class ByteMaskedArray(ByteMaskedMeta[Content], Content):
 
     def _to_backend(self, backend: Backend) -> Self:
         content = self._content.to_backend(backend)
-        mask = self._mask.to_nplike(backend.index_nplike)
+        mask = self._mask.to_nplike(backend.nplike)
         return ByteMaskedArray(
             mask, content, valid_when=self._valid_when, parameters=self._parameters
         )
 
+    def _materialize(self) -> Self:
+        content = self._content._materialize()
+        mask = self._mask.materialize()
+        return ByteMaskedArray(
+            mask, content, valid_when=self._valid_when, parameters=self._parameters
+        )
+
+    @property
+    def _is_all_materialized(self) -> bool:
+        return self._content.is_all_materialized and self._mask.is_all_materialized
+
+    @property
+    def _is_any_materialized(self) -> bool:
+        return self._content.is_any_materialized or self._mask.is_any_materialized
+
     def _is_equal_to(
         self, other: Self, index_dtype: bool, numpyarray: bool, all_parameters: bool
     ) -> bool:
