@@ -15,6 +15,7 @@ from awkward._nplikes.numpy_like import IndexType, NumpyMetadata
 from awkward._nplikes.placeholder import PlaceholderArray
 from awkward._nplikes.shape import ShapeItem, unknown_length
 from awkward._nplikes.typetracer import TypeTracer
+from awkward._nplikes.virtual import VirtualArray, materialize_if_virtual
 from awkward._parameters import (
     parameters_intersect,
     parameters_union,
@@ -129,7 +130,7 @@ class IndexedArray(IndexedMeta[Content], Content):
                 )
             )
 
-        assert index.nplike is content.backend.index_nplike
+        assert index.nplike is content.backend.nplike
 
         self._index = index
         self._content = content
@@ -173,7 +174,7 @@ class IndexedArray(IndexedMeta[Content], Content):
                 inner = content.index
             else:
                 inner = content.to_IndexedOptionArray64().index
-            result = ak.index.Index64.empty(index.length, nplike=backend.index_nplike)
+            result = ak.index.Index64.empty(index.length, nplike=backend.nplike)
 
             backend.maybe_kernel_error(
                 backend[
@@ -233,7 +234,7 @@ class IndexedArray(IndexedMeta[Content], Content):
         assert isinstance(form, self.form_cls)
         key = getkey(self, form, "index")
         container[key] = ak._util.native_to_byteorder(
-            self._index.raw(backend.index_nplike), byteorder
+            self._index.raw(backend.nplike), byteorder
         )
         self._content._to_buffers(form.content, getkey, container, backend, byteorder)
 
@@ -264,7 +265,7 @@ class IndexedArray(IndexedMeta[Content], Content):
 
     def _repr(self, indent, pre, post):
         out = [indent, pre, "<IndexedArray len="]
-        out.append(repr(str(self.length)))
+        out.append(repr(str(ak._util.maybe_length_of(self))))
         out.append(">")
         out.extend(self._repr_extra(indent + "    "))
         out.append("\n")
@@ -293,6 +294,15 @@ class IndexedArray(IndexedMeta[Content], Content):
             return True
         return self._content._is_getitem_at_placeholder()
 
+    def _is_getitem_at_virtual(self) -> bool:
+        is_virtual = (
+            isinstance(self._index.data, VirtualArray)
+            and not self._index.data.is_materialized
+        )
+        if is_virtual:
+            return True
+        return self._content._is_getitem_at_virtual()
+
     def _getitem_at(self, where: IndexType):
         if not self._backend.nplike.known_data:
             self._touch_data(recursive=False)
@@ -351,10 +361,10 @@ class IndexedArray(IndexedMeta[Content], Content):
                 f"cannot fit jagged slice with length {slicestarts.length} into {type(self).__name__} of size {self.length}",
             )
 
-        nextcarry = ak.index.Index64.empty(self.length, self._backend.index_nplike)
+        nextcarry = ak.index.Index64.empty(self.length, self._backend.nplike)
         assert (
-            nextcarry.nplike is self._backend.index_nplike
-            and self._index.nplike is self._backend.index_nplike
+            nextcarry.nplike is self._backend.nplike
+            and self._index.nplike is self._backend.nplike
         )
         self._maybe_index_error(
             self._backend[
@@ -394,12 +404,10 @@ class IndexedArray(IndexedMeta[Content], Content):
         ):
             nexthead, nexttail = ak._slicing.head_tail(tail)
 
-            nextcarry = ak.index.Index64.empty(
-                self._index.length, self._backend.index_nplike
-            )
+            nextcarry = ak.index.Index64.empty(self._index.length, self._backend.nplike)
             assert (
-                nextcarry.nplike is self._backend.index_nplike
-                and self._index.nplike is self._backend.index_nplike
+                nextcarry.nplike is self._backend.nplike
+                and self._index.nplike is self._backend.nplike
             )
             self._maybe_index_error(
                 self._backend[
@@ -442,13 +450,11 @@ class IndexedArray(IndexedMeta[Content], Content):
                 raise ValueError(
                     f"mask length ({mask.length()}) is not equal to {type(self).__name__} length ({self._index.length})"
                 )
-            nextindex = ak.index.Index64.empty(
-                self._index.length, self._backend.index_nplike
-            )
+            nextindex = ak.index.Index64.empty(self._index.length, self._backend.nplike)
             assert (
-                nextindex.nplike is self._backend.index_nplike
-                and mask.nplike is self._backend.index_nplike
-                and self._index.nplike is self._backend.index_nplike
+                nextindex.nplike is self._backend.nplike
+                and mask.nplike is self._backend.nplike
+                and self._index.nplike is self._backend.nplike
             )
             self._backend.maybe_kernel_error(
                 self._backend[
@@ -469,10 +475,10 @@ class IndexedArray(IndexedMeta[Content], Content):
             return next.project()
 
         else:
-            nextcarry = ak.index.Index64.empty(self.length, self._backend.index_nplike)
+            nextcarry = ak.index.Index64.empty(self.length, self._backend.nplike)
             assert (
-                nextcarry.nplike is self._backend.index_nplike
-                and self._index.nplike is self._backend.index_nplike
+                nextcarry.nplike is self._backend.nplike
+                and self._index.nplike is self._backend.nplike
             )
             self._backend.maybe_kernel_error(
                 self._backend[
@@ -556,14 +562,12 @@ class IndexedArray(IndexedMeta[Content], Content):
 
         theirlength = other.length
         mylength = self.length
-        index = ak.index.Index64.empty(
-            (theirlength + mylength), self._backend.index_nplike
-        )
+        index = ak.index.Index64.empty((theirlength + mylength), self._backend.nplike)
 
         content = other._mergemany([self._content])
 
         # Fill `index` with a range starting at zero, up to `theirlength`
-        assert index.nplike is self._backend.index_nplike
+        assert index.nplike is self._backend.nplike
         self._backend.maybe_kernel_error(
             self._backend["awkward_IndexedArray_fill_count", index.dtype.type](
                 index.data,
@@ -574,7 +578,7 @@ class IndexedArray(IndexedMeta[Content], Content):
         )
 
         # Fill remaining indices
-        assert index.nplike is self._backend.index_nplike
+        assert index.nplike is self._backend.nplike
         self._backend.maybe_kernel_error(
             self._backend[
                 "awkward_IndexedArray_fill",
@@ -612,9 +616,7 @@ class IndexedArray(IndexedMeta[Content], Content):
         contents = []
         contentlength_so_far = 0
         length_so_far = 0
-        nextindex = ak.index.Index64.empty(
-            total_length, nplike=self._backend.index_nplike
-        )
+        nextindex = ak.index.Index64.empty(total_length, nplike=self._backend.nplike)
 
         parameters = self._parameters
         for array in head:
@@ -640,8 +642,8 @@ class IndexedArray(IndexedMeta[Content], Content):
                 contents.append(array.content)
                 array_index = array.index
                 assert (
-                    nextindex.nplike is self._backend.index_nplike
-                    and array_index.nplike is self._backend.index_nplike
+                    nextindex.nplike is self._backend.nplike
+                    and array_index.nplike is self._backend.nplike
                 )
                 self._backend.maybe_kernel_error(
                     self._backend[
@@ -660,7 +662,7 @@ class IndexedArray(IndexedMeta[Content], Content):
                 length_so_far += array.length
             else:
                 contents.append(array)
-                assert nextindex.nplike is self._backend.index_nplike
+                assert nextindex.nplike is self._backend.nplike
                 self._backend.maybe_kernel_error(
                     self._backend[
                         "awkward_IndexedArray_fill_count",
@@ -729,17 +731,17 @@ class IndexedArray(IndexedMeta[Content], Content):
             return self.project()._local_index(axis, depth)
 
     def _unique_index(self, index, sorted=True):
-        next = ak.index.Index64.zeros(self.length, nplike=self._backend.index_nplike)
-        length = ak.index.Index64.zeros(1, nplike=self._backend.index_nplike)
+        next = ak.index.Index64.zeros(self.length, nplike=self._backend.nplike)
+        length = ak.index.Index64.zeros(1, nplike=self._backend.nplike)
 
         if not sorted:
             next = self._index
-            offsets = ak.index.Index64.empty(2, self._backend.index_nplike)
+            offsets = ak.index.Index64.empty(2, self._backend.nplike)
             offsets[0] = 0
             offsets[1] = next.length
             assert (
-                next.nplike is self._backend.index_nplike
-                and offsets.nplike is self._backend.index_nplike
+                next.nplike is self._backend.nplike
+                and offsets.nplike is self._backend.nplike
             )
             self._backend.maybe_kernel_error(
                 self._backend[
@@ -760,14 +762,14 @@ class IndexedArray(IndexedMeta[Content], Content):
             )
 
         assert (
-            self._index.nplike is self._backend.index_nplike
-            and next.nplike is self._backend.index_nplike
-            and length.nplike is self._backend.index_nplike
+            self._index.nplike is self._backend.nplike
+            and next.nplike is self._backend.nplike
+            and length.nplike is self._backend.nplike
         )
 
         next = ak.index.Index64(
-            self._backend.index_nplike.unique_values(self._index),
-            nplike=self._backend.index_nplike,
+            self._backend.nplike.unique_values(self._index),
+            nplike=self._backend.nplike,
         )
         length[0] = next.data.size
 
@@ -802,15 +804,15 @@ class IndexedArray(IndexedMeta[Content], Content):
         parents_length = parents.length
         next_length = index_length
 
-        nextcarry = ak.index.Index64.empty(index_length, self._backend.index_nplike)
-        nextparents = ak.index.Index64.empty(index_length, self._backend.index_nplike)
-        outindex = ak.index.Index64.empty(index_length, self._backend.index_nplike)
+        nextcarry = ak.index.Index64.empty(index_length, self._backend.nplike)
+        nextparents = ak.index.Index64.empty(index_length, self._backend.nplike)
+        outindex = ak.index.Index64.empty(index_length, self._backend.nplike)
         assert (
-            nextcarry.nplike is self._backend.index_nplike
-            and nextparents.nplike is self._backend.index_nplike
-            and outindex.nplike is self._backend.index_nplike
-            and self._index.nplike is self._backend.index_nplike
-            and parents.nplike is self._backend.index_nplike
+            nextcarry.nplike is self._backend.nplike
+            and nextparents.nplike is self._backend.nplike
+            and outindex.nplike is self._backend.nplike
+            and self._index.nplike is self._backend.nplike
+            and parents.nplike is self._backend.nplike
         )
         self._backend.maybe_kernel_error(
             self._backend[
@@ -838,14 +840,12 @@ class IndexedArray(IndexedMeta[Content], Content):
         )
 
         if branch or (negaxis is not None and negaxis != depth):
-            nextoutindex = ak.index.Index64.empty(
-                parents_length, self._backend.index_nplike
-            )
+            nextoutindex = ak.index.Index64.empty(parents_length, self._backend.nplike)
             assert (
-                nextoutindex.nplike is self._backend.index_nplike
-                and starts.nplike is self._backend.index_nplike
-                and parents.nplike is self._backend.index_nplike
-                and nextparents.nplike is self._backend.index_nplike
+                nextoutindex.nplike is self._backend.nplike
+                and starts.nplike is self._backend.nplike
+                and parents.nplike is self._backend.nplike
+                and nextparents.nplike is self._backend.nplike
             )
 
             self._backend.maybe_kernel_error(
@@ -883,11 +883,11 @@ class IndexedArray(IndexedMeta[Content], Content):
                     )
 
                 outoffsets = ak.index.Index64.empty(
-                    starts.length + 1, self._backend.index_nplike
+                    starts.length + 1, self._backend.nplike
                 )
                 assert (
-                    outoffsets.nplike is self._backend.index_nplike
-                    and starts.nplike is self._backend.index_nplike
+                    outoffsets.nplike is self._backend.nplike
+                    and starts.nplike is self._backend.nplike
                 )
                 self._backend.maybe_kernel_error(
                     self._backend[
@@ -910,8 +910,8 @@ class IndexedArray(IndexedMeta[Content], Content):
 
             elif isinstance(unique, ak.contents.NumpyArray):
                 nextoutindex = ak.index.Index64(
-                    self._backend.index_nplike.arange(unique.length, dtype=np.int64),
-                    nplike=self._backend.index_nplike,
+                    self._backend.nplike.arange(unique.length, dtype=np.int64),
+                    nplike=self._backend.nplike,
                 )
                 return ak.contents.IndexedOptionArray.simplified(
                     nextoutindex, unique, parameters=self._parameters
@@ -972,7 +972,7 @@ class IndexedArray(IndexedMeta[Content], Content):
         error = self._backend["awkward_IndexedArray_validity", self.index.dtype.type](
             self.index.data, self.index.length, self._content.length, False
         )
-        if error.str is not None:
+        if error is not None and error.str is not None:
             if error.filename is None:
                 filename = ""
             else:
@@ -1018,7 +1018,7 @@ class IndexedArray(IndexedMeta[Content], Content):
             next = IndexedArray(self._index, self._content, parameters=next_parameters)
             return next._to_arrow(pyarrow, mask_node, validbytes, length, options)
 
-        index = self._index.raw(numpy)
+        (index,) = materialize_if_virtual(self._index.raw(numpy))
 
         if self.parameter("__array__") == "categorical":
             dictionary = self._content._to_arrow(
@@ -1089,11 +1089,10 @@ class IndexedArray(IndexedMeta[Content], Content):
             and self._index.length != 0
         ):
             npindex = self._index.data
-            indexmin = self._backend.index_nplike.min(npindex)
-            index = ak.index.Index(
-                npindex - indexmin, nplike=self._backend.index_nplike
-            )
-            content = self._content[indexmin : npindex.max() + 1]
+            indexmin = self._backend.nplike.min(npindex)
+            indexmax = self._backend.nplike.max(npindex)
+            index = ak.index.Index(npindex - indexmin, nplike=self._backend.nplike)
+            content = self._content[indexmin : indexmax + 1]
         else:
             if not self._backend.nplike.known_data:
                 self._touch_data(recursive=False)
@@ -1162,15 +1161,28 @@ class IndexedArray(IndexedMeta[Content], Content):
         if out is not None:
             return out
 
-        index = self._index.raw(numpy)
+        (index,) = materialize_if_virtual(self._index.raw(numpy))
         nextcontent = self._content._carry(ak.index.Index(index), False)
         return nextcontent._to_list(behavior, json_conversions)
 
     def _to_backend(self, backend: Backend) -> Self:
         content = self._content.to_backend(backend)
-        index = self._index.to_nplike(backend.index_nplike)
+        index = self._index.to_nplike(backend.nplike)
         return IndexedArray(index, content, parameters=self._parameters)
 
+    def _materialize(self) -> Self:
+        content = self._content.materialize()
+        index = self._index.materialize()
+        return IndexedArray(index, content, parameters=self._parameters)
+
+    @property
+    def _is_all_materialized(self) -> bool:
+        return self._content.is_all_materialized and self._index.is_all_materialized
+
+    @property
+    def _is_any_materialized(self) -> bool:
+        return self._content.is_any_materialized or self._index.is_any_materialized
+
     def _push_inside_record_or_project(self) -> Self | ak.contents.RecordArray:
         if self.content.is_record:
             return ak.contents.RecordArray(
@@ -1198,7 +1210,7 @@ class IndexedArray(IndexedMeta[Content], Content):
         )
 
     def _trim(self) -> Self:
-        nplike = self._backend.index_nplike
+        nplike = self._backend.nplike
 
         if not nplike.known_data or self._index.length == 0:
             return self
