@@ -3,6 +3,7 @@
 from __future__ import annotations
 
 import math
+from functools import lru_cache, partial
 
 import awkward as ak
 from awkward._backends.dispatch import regularize_backend
@@ -16,6 +17,7 @@ from awkward._nplikes.placeholder import PlaceholderArray
 from awkward._nplikes.shape import ShapeItem, unknown_length
 from awkward._nplikes.virtual import VirtualArray
 from awkward._regularize import is_integer
+from awkward._typing import Callable
 from awkward.forms.form import index_to_dtype, regularize_buffer_key
 
 __all__ = ("from_buffers",)
@@ -153,7 +155,15 @@ def _impl(
     getkey = regularize_buffer_key(buffer_key)
 
     out = _reconstitute(
-        form, length, container, getkey, backend, byteorder, simplify, field_path=()
+        form,
+        length,
+        container,
+        getkey,
+        backend,
+        byteorder,
+        simplify,
+        field_path=(),
+        shape_generator=lambda: (length,),
     )
 
     return wrap_layout(out, highlevel=highlevel, attrs=attrs, behavior=behavior)
@@ -161,22 +171,48 @@ def _impl(
 
 def _from_buffer(
     nplike: NumpyLike,
-    buffer,
+    buffer: Callable | ArrayLike,
     dtype: np.dtype,
     count: ShapeItem,
     byteorder: str,
     field_path: tuple,
+    shape_generator: Callable | None = None,
 ) -> ArrayLike:
-    if callable(buffer):
+    if isinstance(buffer, VirtualArray):
         # This is the case for VirtualArrays
+        # just some checks to make sure the VirtualArray is correctly constructed
+        if nplike != buffer.nplike:
+            raise ValueError(
+                f"Mismatch of nplikes. Got {nplike}, but VirtualArray has {buffer.nplike}."
+            )
+        if dtype != buffer.dtype:
+            raise ValueError(
+                f"Mismatch of dtypes. Got {dtype}, but VirtualArray has {buffer.dtype}."
+            )
+        if count != buffer._shape[0]:
+            raise ValueError(
+                f"Mismatch of lengths. Got {count}, but VirtualArray has {buffer._shape[0]}."
+            )
+        return buffer
+
+    elif callable(buffer):
+        # This is the case where we automatically create VirtualArrays
         # We use recursion here to pass down the from_buffer and byteorder transformations to the generator
+        assert callable(shape_generator), "shape_generator must be callable"
+        cached_shape_generator = lru_cache(maxsize=1)(shape_generator)
+
+        def generator():
+            (length,) = cached_shape_generator()
+            return _from_buffer(
+                nplike, buffer(), dtype, length, byteorder, field_path, None
+            )
+
         return VirtualArray(
             nplike=nplike,
             shape=(count,),
             dtype=dtype,
-            generator=lambda: _from_buffer(
-                nplike, buffer(), dtype, count, byteorder, field_path
-            ),
+            generator=generator,
+            shape_generator=cached_shape_generator,
         )
     # Unknown-length information implies that we didn't load shape-buffers (offsets, etc)
     # for the parent of this node. Thus, this node and its children *must* only
@@ -195,19 +231,30 @@ def _from_buffer(
         copy = None if isinstance(nplike, Jax) else False  # Jax can not avoid this
         array = nplike.reshape(buffer.view(dtype), shape=(-1,), copy=copy)
 
-        if array.size < count:
-            raise TypeError(
-                f"size of array ({array.size}) is less than size of form ({count})"
-            )
-
-        return array[:count]
+        # we can't compare with count or slice when we're working with tracers
+        if not (isinstance(nplike, Jax) and nplike.is_currently_tracing()):
+            if array.size < count:
+                raise TypeError(
+                    f"size of array ({array.size}) is less than size of form ({count})"
+                )
+            return array[:count]
+        else:
+            return array
     else:
         array = nplike.frombuffer(buffer, dtype=dtype, count=count)
         return ak._util.native_to_byteorder(array, byteorder)
 
 
 def _reconstitute(
-    form, length, container, getkey, backend, byteorder, simplify, field_path
+    form,
+    length,
+    container,
+    getkey,
+    backend,
+    byteorder,
+    simplify,
+    field_path,
+    shape_generator,
 ):
     if isinstance(form, ak.forms.EmptyForm):
         if length != 0:
@@ -217,7 +264,16 @@ def _reconstitute(
     elif isinstance(form, ak.forms.NumpyForm):
         dtype = ak.types.numpytype.primitive_to_dtype(form.primitive)
         raw_array = container[getkey(form, "data")]
-        real_length = length * math.prod(form.inner_shape)
+
+        def _adjust_length(length):
+            return length * math.prod(form.inner_shape)
+
+        real_length = _adjust_length(length)
+
+        def _shape_generator():
+            (length,) = shape_generator()
+            return (_adjust_length(length),)
+
         data = _from_buffer(
             backend.nplike,
             raw_array,
@@ -225,6 +281,7 @@ def _reconstitute(
             count=real_length,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=_shape_generator,
         )
         if form.inner_shape != ():
             data = backend.nplike.reshape(data, (length, *form.inner_shape))
@@ -243,6 +300,7 @@ def _reconstitute(
             byteorder,
             simplify,
             field_path,
+            shape_generator,
         )
         if simplify:
             make = ak.contents.UnmaskedArray.simplified
@@ -252,17 +310,27 @@ def _reconstitute(
 
     elif isinstance(form, ak.forms.BitMaskedForm):
         raw_array = container[getkey(form, "mask")]
+
+        def _adjust_length(length):
+            return math.ceil(length / 8.0)
+
+        def _shape_generator():
+            (length,) = shape_generator()
+            return (_adjust_length(length),)
+
         if length is unknown_length:
             next_length = unknown_length
         else:
-            next_length = math.ceil(length / 8.0)
+            next_length = _adjust_length(length)
+
         mask = _from_buffer(
-            backend.index_nplike,
+            backend.nplike,
             raw_array,
             dtype=index_to_dtype[form.mask],
             count=next_length,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=_shape_generator,
         )
         content = _reconstitute(
             form.content,
@@ -273,11 +341,15 @@ def _reconstitute(
             byteorder,
             simplify,
             field_path,
+            _shape_generator,
         )
         if simplify:
             make = ak.contents.BitMaskedArray.simplified
         else:
             make = ak.contents.BitMaskedArray
+        # We need to know the length of a BitMaskedArray to initialize it
+        # as it is an argument in __init__ and is not calculated from the content
+        (length,) = shape_generator()
         return make(
             ak.index.Index(mask),
             content,
@@ -290,12 +362,13 @@ def _reconstitute(
     elif isinstance(form, ak.forms.ByteMaskedForm):
         raw_array = container[getkey(form, "mask")]
         mask = _from_buffer(
-            backend.index_nplike,
+            backend.nplike,
             raw_array,
             dtype=index_to_dtype[form.mask],
             count=length,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=shape_generator,
         )
         content = _reconstitute(
             form.content,
@@ -306,6 +379,7 @@ def _reconstitute(
             byteorder,
             simplify,
             field_path,
+            shape_generator,
         )
         if simplify:
             make = ak.contents.ByteMaskedArray.simplified
@@ -321,22 +395,25 @@ def _reconstitute(
     elif isinstance(form, ak.forms.IndexedOptionForm):
         raw_array = container[getkey(form, "index")]
         index = _from_buffer(
-            backend.index_nplike,
+            backend.nplike,
             raw_array,
             dtype=index_to_dtype[form.index],
             count=length,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=shape_generator,
         )
-        if isinstance(index, PlaceholderArray):
+
+        def _adjust_length(index):
+            return 0 if len(index) == 0 else max(0, backend.nplike.max(index) + 1)
+
+        def _shape_generator():
+            return (_adjust_length(index),)
+
+        if isinstance(index, (PlaceholderArray, VirtualArray)):
             next_length = unknown_length
         else:
-            next_length = (
-                0 if len(index) == 0 else max(0, backend.index_nplike.max(index) + 1)
-            )
-            # free memory
-            if isinstance(index, VirtualArray):
-                index.dematerialize()
+            next_length = _adjust_length(index)
         content = _reconstitute(
             form.content,
             next_length,
@@ -346,6 +423,7 @@ def _reconstitute(
             byteorder,
             simplify,
             field_path,
+            _shape_generator,
         )
         if simplify:
             make = ak.contents.IndexedOptionArray.simplified
@@ -360,26 +438,29 @@ def _reconstitute(
     elif isinstance(form, ak.forms.IndexedForm):
         raw_array = container[getkey(form, "index")]
         index = _from_buffer(
-            backend.index_nplike,
+            backend.nplike,
             raw_array,
             dtype=index_to_dtype[form.index],
             count=length,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=shape_generator,
         )
-        if isinstance(index, PlaceholderArray):
-            next_length = unknown_length
-        else:
-            next_length = (
+
+        def _adjust_length(index):
+            return (
                 0
                 if len(index) == 0
-                else backend.index_nplike.index_as_shape_item(
-                    backend.index_nplike.max(index) + 1
-                )
+                else backend.nplike.index_as_shape_item(backend.nplike.max(index) + 1)
             )
-            # free memory
-            if isinstance(index, VirtualArray):
-                index.dematerialize()
+
+        def _shape_generator():
+            return (_adjust_length(index),)
+
+        if isinstance(index, (PlaceholderArray, VirtualArray)):
+            next_length = unknown_length
+        else:
+            next_length = _adjust_length(index)
         content = _reconstitute(
             form.content,
             next_length,
@@ -389,6 +470,7 @@ def _reconstitute(
             byteorder,
             simplify,
             field_path,
+            _shape_generator,
         )
         if simplify:
             make = ak.contents.IndexedArray.simplified
@@ -404,33 +486,37 @@ def _reconstitute(
         raw_array1 = container[getkey(form, "starts")]
         raw_array2 = container[getkey(form, "stops")]
         starts = _from_buffer(
-            backend.index_nplike,
+            backend.nplike,
             raw_array1,
             dtype=index_to_dtype[form.starts],
             count=length,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=shape_generator,
         )
         stops = _from_buffer(
-            backend.index_nplike,
+            backend.nplike,
             raw_array2,
             dtype=index_to_dtype[form.stops],
             count=length,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=shape_generator,
         )
-        if isinstance(stops, PlaceholderArray):
+
+        def _adjust_length(starts, stops):
+            reduced_stops = stops[starts != stops]
+            return 0 if len(starts) == 0 else backend.nplike.max(reduced_stops)
+
+        def _shape_generator():
+            return (_adjust_length(starts, stops),)
+
+        if isinstance(starts, (PlaceholderArray, VirtualArray)) or isinstance(
+            stops, (PlaceholderArray, VirtualArray)
+        ):
             next_length = unknown_length
         else:
-            reduced_stops = stops[starts != stops]
-            next_length = (
-                0 if len(starts) == 0 else backend.index_nplike.max(reduced_stops)
-            )
-            # free memory
-            if isinstance(starts, VirtualArray):
-                starts.dematerialize()
-            if isinstance(stops, VirtualArray):
-                stops.dematerialize()
+            next_length = _adjust_length(starts, stops)
         content = _reconstitute(
             form.content,
             next_length,
@@ -440,6 +526,7 @@ def _reconstitute(
             byteorder,
             simplify,
             field_path,
+            _shape_generator,
         )
         return ak.contents.ListArray(
             ak.index.Index(starts),
@@ -450,22 +537,33 @@ def _reconstitute(
 
     elif isinstance(form, ak.forms.ListOffsetForm):
         raw_array = container[getkey(form, "offsets")]
+
+        def _shape_generator():
+            (first,) = shape_generator()
+            return (first + 1,)
+
         offsets = _from_buffer(
-            backend.index_nplike,
+            backend.nplike,
             raw_array,
             dtype=index_to_dtype[form.offsets],
             count=length + 1,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=_shape_generator,
         )
 
-        if isinstance(offsets, PlaceholderArray):
+        # next length
+        def _adjust_length(offsets):
+            return 0 if len(offsets) == 1 else offsets[-1]
+
+        def _shape_generator():
+            return (_adjust_length(offsets),)
+
+        if isinstance(offsets, (PlaceholderArray, VirtualArray)):
             next_length = unknown_length
         else:
-            next_length = 0 if len(offsets) == 1 else offsets[-1]
-            # free memory
-            if isinstance(offsets, VirtualArray):
-                offsets.dematerialize()
+            next_length = _adjust_length(offsets)
+
         content = _reconstitute(
             form.content,
             next_length,
@@ -475,6 +573,7 @@ def _reconstitute(
             byteorder,
             simplify,
             field_path,
+            _shape_generator,
         )
         return ak.contents.ListOffsetArray(
             ak.index.Index(offsets),
@@ -483,7 +582,16 @@ def _reconstitute(
         )
 
     elif isinstance(form, ak.forms.RegularForm):
-        next_length = length * form.size
+
+        def _adjust_length(length):
+            return length * form.size
+
+        next_length = _adjust_length(length)
+
+        def _shape_generator():
+            (first,) = shape_generator()
+            return (_adjust_length(first),)
+
         content = _reconstitute(
             form.content,
             next_length,
@@ -493,6 +601,7 @@ def _reconstitute(
             byteorder,
             simplify,
             field_path,
+            _shape_generator,
         )
         return ak.contents.RegularArray(
             content,
@@ -512,6 +621,7 @@ def _reconstitute(
                 byteorder,
                 simplify,
                 (*field_path, field),
+                shape_generator,
             )
             for content, field in zip(form.contents, form.fields)
         ]
@@ -526,36 +636,48 @@ def _reconstitute(
         raw_array1 = container[getkey(form, "tags")]
         raw_array2 = container[getkey(form, "index")]
         tags = _from_buffer(
-            backend.index_nplike,
+            backend.nplike,
             raw_array1,
             dtype=index_to_dtype[form.tags],
             count=length,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=shape_generator,
         )
         index = _from_buffer(
-            backend.index_nplike,
+            backend.nplike,
             raw_array2,
             dtype=index_to_dtype[form.index],
             count=length,
             byteorder=byteorder,
             field_path=field_path,
+            shape_generator=shape_generator,
         )
-        if isinstance(index, PlaceholderArray) or isinstance(tags, PlaceholderArray):
+
+        def _adjust_length(index, tags, tag):
+            selected_index = index[tags == tag]
+            if len(selected_index) == 0:
+                return 0
+            else:
+                return backend.nplike.max(selected_index) + 1
+
+        _shape_generators = []
+        for tag in range(len(form.contents)):
+
+            def _shape_generator(tag):
+                return (_adjust_length(index, tags, tag),)
+
+            _shape_generators.append(partial(_shape_generator, tag=tag))
+
+        if isinstance(index, (PlaceholderArray, VirtualArray)) or isinstance(
+            tags, (PlaceholderArray, VirtualArray)
+        ):
             lengths = [unknown_length] * len(form.contents)
         else:
             lengths = []
             for tag in range(len(form.contents)):
-                selected_index = index[tags == tag]
-                if len(selected_index) == 0:
-                    lengths.append(0)
-                else:
-                    lengths.append(backend.index_nplike.max(selected_index) + 1)
-            # free memory
-            if isinstance(index, VirtualArray):
-                index.dematerialize()
-            if isinstance(tags, VirtualArray):
-                tags.dematerialize()
+                lengths.append(_adjust_length(index, tags, tag))
+
         contents = [
             _reconstitute(
                 content,
@@ -566,6 +688,7 @@ def _reconstitute(
                 byteorder,
                 simplify,
                 field_path,
+                _shape_generators[i],
             )
             for i, content in enumerate(form.contents)
         ]
